import { DiagnosticTarget, NoTarget, Program, Scalar } from "@typespec/compiler";
import { JsContext } from "../ctx.js";
/**
 * Emits a declaration for a scalar type.
 *
 * This is rare in TypeScript, as the scalar will ordinarily be used inline, but may be desirable in some cases.
 *
 * @param ctx - The emitter context.
 * @param scalar - The scalar to emit.
 * @returns a string that declares an alias to the scalar type in TypeScript.
 */
export declare function emitScalar(ctx: JsContext, scalar: Scalar): string;
/**
 * Get the string parsing template for a given scalar.
 *
 * It is common that a scalar type is encoded as a string. For example, in HTTP path parameters or query parameters
 * where the value may be an integer, but the APIs expose it as a string. In such cases the parse template may be
 * used to coerce the string value to the correct scalar type.
 *
 * The result of this function contains the string "{}" exactly once, which should be replaced with the text of an
 * expression evaluating to the string representation of the scalar.
 *
 * For example, scalars that are represented by JS `number` are parsed with the template `Number({})`, which will
 * convert the string to a number.
 *
 * @param ctx - The emitter context.
 * @param scalar - The scalar to parse from a string
 * @returns a template expression string that can be used to parse a string into the scalar type.
 */
export declare function parseTemplateForScalar(ctx: JsContext, scalar: Scalar): string;
/**
 * Get the string encoding template for a given scalar.
 * @param ctx
 * @param scalar
 */
export declare function encodeTemplateForScalar(ctx: JsContext, scalar: Scalar): string;
/**
 * Gets a TypeScript type that can represent a given TypeSpec scalar.
 *
 * Scalar recognition is recursive. If a scalar is not recognized, we will treat it as its parent scalar and try again.
 *
 * If no scalar in the chain is recognized, it will be treated as `unknown` and a warning will be issued.
 *
 * @param program - The program that contains the scalar
 * @param scalar - The scalar to get the TypeScript type for
 * @param diagnosticTarget - Where to report a diagnostic if the scalar is not recognized.
 * @returns a string containing a TypeScript type that can represent the scalar
 */
export declare function getJsScalar(program: Program, scalar: Scalar, diagnosticTarget: DiagnosticTarget | typeof NoTarget): string;
//# sourceMappingURL=scalar.d.ts.map